import pickle
from itertools import combinations
from unittest.mock import Mock

import numpy as np
import pytest

import mici


def _bind_with_decorator(instance, name, func, decorator):
    """Bind func decorated with decorator as a method of instance and return.

    Based on https://stackoverflow.com/a/1015405
    """
    func.__name__ = name
    decorated_func = decorator(func)
    bound_method = decorated_func.__get__(instance, instance.__class__)
    setattr(instance, name, bound_method)
    return bound_method


@pytest.fixture
def state_vars():
    return {"spam": np.array([0.5, -1.0]), "ham": np.array(1), "eggs": -2.0}


def test_state_construction(state_vars):
    state = mici.states.ChainState(**state_vars)
    for key, val in state_vars.items():
        assert hasattr(state, key)
        assert getattr(state, key) is val


def test_state_copy(state_vars):
    state = mici.states.ChainState(**state_vars)
    state_copy = state.copy()
    for key, val in state_vars.items():
        assert hasattr(state_copy, key)
        assert np.all(getattr(state_copy, key) == val)


def test_state_copy_independent(state_vars):
    state = mici.states.ChainState(**state_vars)
    state_copy = state.copy()
    for key, val in state_vars.items():
        attr = getattr(state_copy, key)
        attr *= 2
        assert np.all(getattr(state, key) == val)


def test_state_copy_call_counts_same_obj(state_vars):
    state = mici.states.ChainState(**state_vars, _call_counts={})
    state_copy = state.copy()
    assert state._call_counts is state_copy._call_counts


def test_state_contains(state_vars):
    state = mici.states.ChainState(**state_vars)
    for key in state_vars:
        assert key in state


def test_state_read_only(state_vars):
    state = mici.states.ChainState(**state_vars, _read_only=True)
    for key in state_vars:
        with pytest.raises(mici.errors.ReadOnlyStateError):
            setattr(state, key, None)


def test_state_copy_read_only(state_vars):
    state = mici.states.ChainState(**state_vars)
    state_copy = state.copy(read_only=True)
    for key in state_vars:
        with pytest.raises(mici.errors.ReadOnlyStateError):
            setattr(state_copy, key, None)


def test_state_pickling(state_vars):
    state = mici.states.ChainState(**state_vars)
    pickled_state = pickle.dumps(state)
    unpickled_state = pickle.loads(pickled_state)  # noqa: S301
    assert isinstance(unpickled_state, mici.states.ChainState)
    for key, val in state_vars.items():
        assert hasattr(unpickled_state, key)
        assert np.all(getattr(unpickled_state, key) == val)


def test_state_to_string(state_vars):
    state = mici.states.ChainState(**state_vars)
    assert isinstance(str(state), str)


def test_state_representation(state_vars):
    state = mici.states.ChainState(**state_vars)
    assert isinstance(repr(state), str)


def _mock_memoized_one_output_system_methods(system, state_vars):
    for var_name in state_vars:
        name = f"{var_name}_method"

        def func(state, var_name=var_name):
            return 2 * getattr(state, var_name)

        mocked_method = Mock(wraps=lambda _self, state: func(state), name=name)
        bound_memoized_method = _bind_with_decorator(
            system,
            name,
            mocked_method,
            mici.states.cache_in_state(var_name),
        )
        yield (var_name,), func, mocked_method, bound_memoized_method
    for var_names in combinations(state_vars.keys(), 2):
        name = f"{var_names[0]}_{var_names[1]}_method"

        def func(state, var_names=var_names):
            return getattr(state, var_names[0]) + getattr(state, var_names[1])

        mocked_method = Mock(wraps=lambda _self, state: func(state), name=name)
        bound_memoized_method = _bind_with_decorator(
            system,
            name,
            mocked_method,
            mici.states.cache_in_state(*var_names),
        )
        yield var_names, func, mocked_method, bound_memoized_method


def test_cache_in_state(state_vars):
    system = Mock(name="MockSystem")
    state = mici.states.ChainState(**state_vars, _call_counts={})
    call_counts = state._call_counts
    for (
        var_names,
        func,
        mocked_method,
        bound_memoized_method,
    ) in _mock_memoized_one_output_system_methods(system, state_vars):
        assert (
            mocked_method.call_count == 0
        ), "method to be memoized should not be called by decorator"
        ret_val_1 = bound_memoized_method(state)
        assert np.all(
            ret_val_1 == func(state),
        ), "return value of memoized method should be same as func"
        assert (
            mocked_method.call_count == 1
        ), "memoized method should be executed once on initial call"
        cache_key = mici.states._cache_key_func(system, mocked_method)
        assert (
            call_counts[cache_key] == mocked_method.call_count
        ), f"state call counts value for {cache_key} incorrect"
        assert cache_key in state._cache, f"state cache dict should contain {cache_key}"
        assert (
            state._cache[cache_key] is ret_val_1
        ), f"cached value for key {cache_key} is incorrect"
        ret_val_2 = bound_memoized_method(state)
        assert (
            mocked_method.call_count == 1
        ), "memoized method should not be executed again on second call"
        assert (
            call_counts[cache_key] == mocked_method.call_count
        ), f"state call counts value for {cache_key} incorrect"
        assert (
            ret_val_1 is ret_val_2
        ), "state cache should return same value when state unchanged"
        for var_name in var_names:
            setattr(state, var_name, state_vars[var_name])
            assert (
                state._cache[cache_key] is None
            ), f"cached value for key {cache_key} should be None"
        ret_val_3 = bound_memoized_method(state)
        assert (
            mocked_method.call_count == 2  # noqa: PLR2004
        ), f"memoized method should be recalled after {var_names} update"
        assert (
            call_counts[cache_key] == mocked_method.call_count
        ), f"state call counts value for {cache_key} incorrect"
        assert np.all(
            ret_val_3 == func(state),
        ), "return value of memoized method should be same as unmemoized"


def _mock_memoized_aux_output_system_methods(system, state_vars):
    for var_name in state_vars:
        name = f"{var_name}_method"

        def aux_func(state, var_name=var_name):
            return -1 * getattr(state, var_name)

        def prim_func(state, var_name=var_name):
            return 2 * getattr(state, var_name)

        mocked_method = Mock(
            wraps=lambda _self, state: (prim_func(state), aux_func(state)),
            name=name,
        )
        mocked_aux_method = Mock(
            wraps=lambda _self, state: aux_func(state),
            name="aux_" + name,
        )
        bound_memoized_method = _bind_with_decorator(
            system,
            name,
            mocked_method,
            mici.states.cache_in_state_with_aux(var_name, "aux_" + name),
        )
        bound_memoized_aux_method = _bind_with_decorator(
            system,
            "aux_" + name,
            mocked_aux_method,
            mici.states.cache_in_state(var_name),
        )
        yield (
            (var_name,),
            prim_func,
            aux_func,
            mocked_method,
            mocked_aux_method,
            bound_memoized_method,
            bound_memoized_aux_method,
        )
    for var_names in combinations(state_vars.keys(), 2):
        name = f"{var_names[0]}_{var_names[1]}_method"

        def prim_func(state, var_names=var_names):
            return getattr(state, var_names[0]) + getattr(state, var_names[1])

        def aux_func(state, var_names=var_names):
            return -1 * getattr(state, var_names[0])

        mocked_method = Mock(
            wraps=lambda _self, state: (prim_func(state), aux_func(state)),
            name=name,
        )
        mocked_aux_method = Mock(
            wraps=lambda _self, state: aux_func(state),
            name="aux_" + name,
        )
        bound_memoized_method = _bind_with_decorator(
            system,
            name,
            mocked_method,
            mici.states.cache_in_state_with_aux(var_names, "aux_" + name),
        )
        bound_memoized_aux_method = _bind_with_decorator(
            system,
            "aux_" + name,
            mocked_aux_method,
            mici.states.cache_in_state(*var_name),
        )
        yield (
            var_names,
            prim_func,
            aux_func,
            mocked_method,
            mocked_aux_method,
            bound_memoized_method,
            bound_memoized_aux_method,
        )


def test_cache_in_state_with_aux(state_vars):
    system = Mock(name="MockSystem")
    state = mici.states.ChainState(**state_vars, _call_counts={})
    call_counts = state._call_counts
    for (
        var_names,
        prim_func,
        aux_func,
        mocked_method,
        mocked_aux_method,
        bound_memoized_method,
        bound_memoized_aux_method,
    ) in _mock_memoized_aux_output_system_methods(system, state_vars):
        assert (
            mocked_method.call_count == 0
        ), "method to be memoized should not be called by decorator"
        ret_val_1 = bound_memoized_method(state)
        assert np.all(
            ret_val_1 == prim_func(state),
        ), "return value of memoized method should be same as prim_func"
        assert (
            mocked_method.call_count == 1
        ), "memoized method should be executed once on first call"
        assert (
            mocked_aux_method.call_count == 0
        ), "memoized aux method should not have been executed"
        prim_cache_key = mici.states._cache_key_func(system, mocked_method)
        aux_cache_key = mici.states._cache_key_func(system, mocked_aux_method)
        assert (
            call_counts[prim_cache_key] == mocked_method.call_count
        ), f"state call counts value for {prim_cache_key} incorrect"
        assert (
            call_counts[aux_cache_key] == mocked_aux_method.call_count
        ), f"state call counts value for {aux_cache_key} incorrect"
        assert (
            prim_cache_key in state._cache
        ), f"state cache dict should contain {prim_cache_key}"
        assert (
            aux_cache_key in state._cache
        ), f"state cache dict should contain {aux_cache_key}"
        assert (
            state._cache[prim_cache_key] is ret_val_1
        ), f"cached value for key {prim_cache_key} is incorrect"
        assert np.all(
            state._cache[aux_cache_key] == aux_func(state),
        ), f"cached value for key {aux_cache_key} is incorrect"
        aux_ret_val = bound_memoized_aux_method(state)
        assert (
            mocked_aux_method.call_count == 0
        ), "memoized aux method should not have been executed"
        assert (
            call_counts[aux_cache_key] == mocked_aux_method.call_count
        ), f"state call counts value for {aux_cache_key} incorrect"
        assert np.all(
            aux_ret_val == aux_func(state),
        ), "return value of aux memoized method is incorrect"
        ret_val_2 = bound_memoized_method(state)
        assert (
            mocked_method.call_count == 1
        ), "memoized method should not be executed again on second call"
        assert (
            call_counts[prim_cache_key] == mocked_method.call_count
        ), f"state call counts value for {prim_cache_key} incorrect"
        assert (
            call_counts[aux_cache_key] == mocked_aux_method.call_count
        ), f"state call counts value for {aux_cache_key} incorrect"
        assert (
            ret_val_1 is ret_val_2
        ), "state cache should return same value when state unchanged"
        for var_name in var_names:
            setattr(state, var_name, state_vars[var_name])
            assert (
                state._cache[prim_cache_key] is None
            ), f"cached value for key {prim_cache_key} should be None"
            assert (
                state._cache[aux_cache_key] is None
            ), f"cached value for key {aux_cache_key} should be None"
        ret_val_3 = bound_memoized_method(state)
        assert (
            mocked_method.call_count == 2  # noqa: PLR2004
        ), f"memoized method should be recalled after {var_names} update"
        assert (
            call_counts[prim_cache_key] == mocked_method.call_count
        ), f"state call counts value for {prim_cache_key} incorrect"
        assert np.all(
            state._cache[aux_cache_key] == aux_func(state),
        ), f"cached value for key {aux_cache_key} is incorrect"
        assert np.all(
            ret_val_3 == prim_func(state),
        ), "return value of memoized method should be same as unmemoized"



from collections import OrderedDict
from queue import SimpleQueue

import pytest

import mici


def test_format_time():
    assert mici.progressbars._format_time(100) == "01:40"
    assert mici.progressbars._format_time(10000) == "2:46:40"


def test_stats_runnin_means():
    means = {}
    vals = {"a": 1, "b": 2}
    mici.progressbars._update_stats_running_means(1, means, vals)
    assert means == vals
    mici.progressbars._update_stats_running_means(2, means, vals)
    assert means == vals


@pytest.fixture(
    params=(
        "DummyProgressBar",
        "SequenceProgressBar",
        "LabelledSequenceProgressBar",
        "_ProxySequenceProgressBar",
    ),
)
def progress_bar_and_sequence(request):
    if request.param == "LabelledSequenceProgressBar":
        sequence = OrderedDict()
        sequence["a"] = 1
        sequence["b"] = 2
        sequence["c"] = 3
        return (
            mici.progressbars.LabelledSequenceProgressBar(sequence),
            sequence.values(),
        )
    sequence = range(100)
    if request.param == "DummyProgressBar":
        return mici.progressbars.DummyProgressBar(sequence, None), sequence
    if request.param == "SequenceProgressBar":
        return mici.progressbars.SequenceProgressBar(sequence, None), sequence
    if request.param == "_ProxySequenceProgressBar":
        return (
            mici.progressbars._ProxySequenceProgressBar(sequence, 0, SimpleQueue()),
            sequence,
        )
    return None


def test_progress_bar_len(progress_bar_and_sequence):
    progress_bar, sequence = progress_bar_and_sequence
    assert len(progress_bar) == len(sequence)


def test_progress_bar_iter(progress_bar_and_sequence):
    progress_bar, sequence = progress_bar_and_sequence
    with progress_bar:
        for (val, iter_dict), val_orig in zip(progress_bar, sequence, strict=True):
            assert val == val_orig
            assert isinstance(iter_dict, dict)



import numpy as np
import pytest

import mici

SEED = 3046987125
STATE_DIM = 2
N_CHAIN = 2


@pytest.fixture
def rng():
    return np.random.default_rng(SEED)


def neg_log_dens(pos):
    return np.sum(pos**2) / 2


def grad_neg_log_dens(pos):
    return pos


@pytest.fixture
def system():
    return mici.systems.EuclideanMetricSystem(
        neg_log_dens=neg_log_dens,
        grad_neg_log_dens=grad_neg_log_dens,
    )


@pytest.fixture
def integrator(system):
    return mici.integrators.LeapfrogIntegrator(system, step_size=0.5)


@pytest.fixture(params=("dict", "list", "tuple", "nested"))
def memmap_obj_and_filename_obj(request, tmp_path):
    filename = tmp_path / "mm.npy"
    mm = np.lib.format.open_memmap(filename, dtype=np.float64, mode="w+", shape=(1,))
    mm[:] = np.arange(mm.size)
    if request.param == "dict":

        def pytree_func(obj):
            return {"k": obj}

    elif request.param == "list":

        def pytree_func(obj):
            return [obj]

    elif request.param == "tuple":

        def pytree_func(obj):
            return (obj,)

    elif request.param == "nested":

        def pytree_func(obj):
            return {"k": [(obj,)]}

    else:
        msg = f"Invalid param {request.param}"
        raise ValueError(msg)
    return pytree_func(mm), pytree_func(filename)


def test_memmaps_to_file_paths(memmap_obj_and_filename_obj):
    memmap_obj, filename_obj = memmap_obj_and_filename_obj
    generated_filename_obj = mici.samplers._memmaps_to_file_paths(memmap_obj)
    assert filename_obj == generated_filename_obj


def test_file_paths_to_memmaps(memmap_obj_and_filename_obj):
    memmap_obj, filename_obj = memmap_obj_and_filename_obj
    generated_memmap_obj = mici.samplers._file_paths_to_memmaps(filename_obj)
    assert memmap_obj == generated_memmap_obj


def test_zip_dict():
    assert list(mici.samplers._zip_dict(a=[1, 2, 3], b=[4, 5, 6])) == [
        {"a": 1, "b": 4},
        {"a": 2, "b": 5},
        {"a": 3, "b": 6},
    ]


@pytest.fixture(params=("Generator(PCG64)", "Generator(SFC64)", "RandomState"))
def base_rng(request):
    if request.param == "Generator(PCG64)":
        return np.random.Generator(np.random.PCG64(SEED))
    if request.param == "Generator(SFC64)":
        return np.random.Generator(np.random.SFC64(SEED))
    return np.random.RandomState(SEED)


def test_get_per_chain_rngs(base_rng):
    rngs = mici.samplers._get_per_chain_rngs(base_rng, 4)
    for i, rng in enumerate(rngs):
        assert isinstance(rng, np.random.Generator)
        if i != 0:
            assert rng is not rngs[i - 1]


def test_get_per_chain_rngs_raises():
    with pytest.raises(ValueError, match="Unsupported random number generator"):
        mici.samplers._get_per_chain_rngs(None, 4)


def test_random_state_raises_deprecation_warning():
    rng = np.random.RandomState(SEED)
    with pytest.deprecated_call():
        mici.samplers.MarkovChainMonteCarloMethod(rng=rng, transitions={})


class MarkovChainMonteCarloMethodTests:
    def test_sampler_attributes(self, sampler, rng):
        assert sampler.rng is rng
        assert isinstance(sampler.transitions, dict)

    @pytest.mark.parametrize("n_warm_up_iter", [0, 2])
    @pytest.mark.parametrize("n_main_iter", [0, 2])
    @pytest.mark.parametrize("trace_warm_up", [True, False])
    @pytest.mark.parametrize("n_process", [1, N_CHAIN])
    def test_sample_chains(
        self,
        sampler,
        n_warm_up_iter,
        n_main_iter,
        init_states,
        trace_funcs,
        trace_warm_up,
        n_process,
        kwargs,
    ):
        final_states, traces, stats = sampler.sample_chains(
            n_warm_up_iter=n_warm_up_iter,
            n_main_iter=n_main_iter,
            init_states=init_states,
            trace_funcs=trace_funcs,
            trace_warm_up=trace_warm_up,
            n_process=n_process,
            **kwargs,
        )
        trace_vars = {}
        if trace_funcs is not None:
            for trace_func in trace_funcs:
                trace_vars.update(trace_func(final_states[0]))
        assert all(isinstance(state, mici.states.ChainState) for state in final_states)
        n_trace_iter = n_main_iter + n_warm_up_iter if trace_warm_up else n_main_iter
        if trace_funcs is None or len(trace_funcs) == 0:
            assert traces is None
        else:
            assert isinstance(traces, dict)
            assert traces.keys() == trace_vars.keys()
            self.check_traces(traces, n_trace_iter, len(init_states), trace_vars)
        assert isinstance(stats, dict)
        self.check_stats_dict(
            stats,
            n_trace_iter,
            len(init_states),
            sampler.transitions,
        )

    def check_traces(self, traces, n_iter, n_chain, trace_vars):
        for trace_key, trace_array_list in traces.items():
            assert len(trace_array_list) == n_chain
            for trace_array in trace_array_list:
                assert isinstance(trace_array, np.ndarray)
                assert not np.any(np.isnan(trace_array))
                assert trace_array.shape[0] == n_iter
                assert trace_array.shape[1:] == np.shape(trace_vars[trace_key])

    def check_trans_stats_dict(self, trans_stats, n_iter, statistic_types):
        for stat_key, stat_array_list in trans_stats.items():
            for stat_array in stat_array_list:
                assert stat_array.shape[0] == n_iter
                assert not np.any(np.isnan(stat_array))
                assert stat_array.dtype == statistic_types[stat_key][0]

    def check_stats_dict(self, stats, n_iter, n_chain, transitions):
        for trans_key, trans_stats in stats.items():
            assert isinstance(trans_key, str)
            assert trans_key in transitions
            statistic_types = transitions[trans_key].statistic_types
            if statistic_types is None:
                assert len(trans_stats) == 0
            else:
                assert trans_stats.keys() == statistic_types.keys()
                self.check_trans_stats_dict(trans_stats, n_iter, statistic_types)


class TestMarkovChainMonteCarloMethod(MarkovChainMonteCarloMethodTests):
    @staticmethod
    def pos_mom_trace_func(state):
        return {"pos": state.pos, "mom": state.mom}

    @pytest.fixture(params=("pos_mom_trace_func", "empty", None))
    def trace_funcs(self, request):
        if request.param == "pos_mom_trace_func":
            return [self.pos_mom_trace_func]
        if request.param == "empty":
            return []
        if request.param is None:
            return None
        msg = f"Invalid param {request.param}"
        raise ValueError(msg)

    @pytest.fixture(params=(None, (("integration", "accept_stat"),)))
    def monitor_stats(self, request):
        return request.param

    @pytest.fixture(params=(None, "empty", "DualAveragingStepSizeAdapter"))
    def adapters(self, request):
        if request.param is None:
            return None
        if request.param == "empty":
            return {}
        if request.param == "DualAveragingStepSizeAdapter":
            return {"integration": [mici.adapters.DualAveragingStepSizeAdapter()]}
        msg = f"Invalid param {request.param}"
        raise ValueError(msg)

    @pytest.fixture(params=(None, "WarmUpStager"))
    def stager(self, request):
        if request.param is None:
            return None
        if request.param == "WarmUpStager":
            return mici.stagers.WarmUpStager()
        msg = f"Invalid param {request.param}"
        raise ValueError(msg)

    @pytest.fixture(params=((True, None), (True, "tmp"), (False, None)))
    def force_memmap_and_memmap_path(self, request, tmp_path_factory):
        force_memmap, memmap_path = request.param
        if memmap_path == "tmp":
            memmap_path = tmp_path_factory.mktemp("traces-and-stats")
        return force_memmap, memmap_path

    @pytest.fixture(params=(True, False))
    def display_progress(self, request):
        return request.param

    @pytest.fixture
    def kwargs(
        self,
        monitor_stats,
        adapters,
        stager,
        force_memmap_and_memmap_path,
        display_progress,
    ):
        force_memmap, memmap_path = force_memmap_and_memmap_path
        return {
            "monitor_stats": monitor_stats,
            "adapters": adapters,
            "stager": stager,
            "force_memmap": force_memmap,
            "memmap_path": memmap_path,
            "display_progress": display_progress,
        }

    @pytest.fixture
    def sampler(self, integrator, system, rng):
        return mici.samplers.MarkovChainMonteCarloMethod(
            rng=rng,
            transitions={
                "momentum": mici.transitions.CorrelatedMomentumTransition(system, 0.5),
                "integration": mici.transitions.MetropolisStaticIntegrationTransition(
                    system,
                    integrator,
                    1,
                ),
            },
        )

    @pytest.fixture(params=("ChainState", "dict"))
    def init_states(self, rng, request):
        init_states = []
        for _c in range(N_CHAIN):
            pos, mom = rng.standard_normal((2, STATE_DIM))
            if request.param == "dict":
                state = {"pos": pos, "mom": mom, "dir": 1}
            elif request.param == "ChainState":
                state = mici.states.ChainState(pos=pos, mom=mom, dir=1)
            else:
                msg = f"Invalid param {request.param}"
                raise ValueError(msg)
            init_states.append(state)
        return init_states


class HamiltonianMCMCTests(MarkovChainMonteCarloMethodTests):
    @pytest.fixture(params=(None, ("accept_stat", "n_step")))
    def monitor_stats(self, request):
        return request.param

    @pytest.fixture(params=(None, "DualAveragingStepSizeAdapter"))
    def adapters(self, request):
        if request.param is None:
            return None
        if request.param == "DualAveragingStepSizeAdapter":
            return [mici.adapters.DualAveragingStepSizeAdapter()]
        msg = f"Invalid param {request.param}"
        raise ValueError(msg)

    @pytest.fixture
    def kwargs(self, monitor_stats, adapters):
        kwargs = {}
        if monitor_stats is not None:
            kwargs["monitor_stats"] = monitor_stats
        if adapters is not None:
            kwargs["adapters"] = adapters
        return kwargs

    @pytest.fixture
    def trace_funcs(self, sampler):
        return [sampler._default_trace_func]

    @pytest.fixture(params=("ChainState_with_mom", "ChainState_no_mom", "array"))
    def init_states(self, rng, request):
        init_states = []
        for _c in range(N_CHAIN):
            pos = rng.standard_normal(STATE_DIM)
            if request.param == "array":
                state = pos
            elif request.param == "ChainState_no_mom":
                state = mici.states.ChainState(pos=pos, mom=None, dir=1)
            elif request.param == "ChainState_with_mom":
                mom = rng.standard_normal(STATE_DIM)
                state = mici.states.ChainState(pos=pos, mom=mom, dir=1)
            else:
                msg = f"Invalid param {request.param}"
                raise ValueError(msg)
            init_states.append(state)
        return init_states

    def check_stats_dict(self, stats, n_iter, n_chain, transitions):
        statistic_types = transitions["integration_transition"].statistic_types
        self.check_trans_stats_dict(stats, n_iter, statistic_types)


class TestStaticMetropolisHMC(HamiltonianMCMCTests):
    n_step = 2

    def test_max_tree_depth(self, sampler):
        assert sampler.n_step == self.n_step
        new_n_step = self.n_step + 1
        sampler.n_step = new_n_step
        assert sampler.n_step == new_n_step
        assert sampler.transitions["integration_transition"].n_step == (new_n_step)

    @pytest.fixture
    def sampler(self, integrator, system, rng):
        return mici.samplers.StaticMetropolisHMC(
            system=system,
            integrator=integrator,
            rng=rng,
            n_step=self.n_step,
        )


class TestRandomMetropolisHMC(HamiltonianMCMCTests):
    n_step_range = (1, 3)

    def test_max_tree_depth(self, sampler):
        assert sampler.n_step_range == self.n_step_range
        new_n_step_range = (self.n_step_range[0] + 1, self.n_step_range[1] + 1)
        sampler.n_step_range = new_n_step_range
        assert sampler.n_step_range == new_n_step_range
        assert sampler.transitions["integration_transition"].n_step_range == (
            new_n_step_range
        )

    @pytest.fixture
    def sampler(self, integrator, system, rng):
        return mici.samplers.RandomMetropolisHMC(
            system=system,
            integrator=integrator,
            rng=rng,
            n_step_range=self.n_step_range,
        )


class DynamicHMCTests(HamiltonianMCMCTests):
    max_tree_depth = 2
    max_delta_h = 1000

    def test_max_tree_depth(self, sampler):
        assert sampler.max_tree_depth == self.max_tree_depth
        new_max_tree_depth = self.max_tree_depth + 1
        sampler.max_tree_depth = new_max_tree_depth
        assert sampler.max_tree_depth == new_max_tree_depth
        assert sampler.transitions["integration_transition"].max_tree_depth == (
            new_max_tree_depth
        )

    def test_max_delta_h(self, sampler):
        assert sampler.max_delta_h == self.max_delta_h
        new_max_delta_h = self.max_delta_h + 1
        sampler.max_delta_h = new_max_delta_h
        assert sampler.max_delta_h == new_max_delta_h
        assert sampler.transitions["integration_transition"].max_delta_h == (
            new_max_delta_h
        )


class TestDynamicMultinomialHMC(DynamicHMCTests):
    @pytest.fixture
    def sampler(self, integrator, system, rng):
        return mici.samplers.DynamicMultinomialHMC(
            system=system,
            integrator=integrator,
            rng=rng,
            max_tree_depth=self.max_tree_depth,
            max_delta_h=self.max_delta_h,
        )


class TestDynamicDynamicSliceHMC(DynamicHMCTests):
    @pytest.fixture
    def sampler(self, integrator, system, rng):
        return mici.samplers.DynamicSliceHMC(
            system=system,
            integrator=integrator,
            rng=rng,
            max_tree_depth=self.max_tree_depth,
            max_delta_h=self.max_delta_h,
        )



"""Autograd differential operators."""

from __future__ import annotations

from typing import TYPE_CHECKING

AUTOGRAD_AVAILABLE = True
try:
    import autograd.numpy as np
    from autograd.builtins import tuple as atuple
    from autograd.core import make_vjp
    from autograd.extend import vspace
except ImportError:
    AUTOGRAD_AVAILABLE = False

if TYPE_CHECKING:
    from mici.types import (
        ArrayFunction,
        ArrayLike,
        GradientFunction,
        HessianFunction,
        JacobianFunction,
        MatrixHessianProduct,
        MatrixHessianProductFunction,
        MatrixTressianProduct,
        MatrixTressianProductFunction,
        ScalarFunction,
        ScalarLike,
        VectorJacobianProduct,
        VectorJacobianProductFunction,
    )


def grad_and_value(func: ScalarFunction) -> GradientFunction:
    """Makes a function that returns both gradient and value of a function."""

    def grad_and_value_func(x: ArrayLike) -> tuple[ArrayLike, ScalarLike]:
        vjp, val = make_vjp(func, x)
        if vspace(val).size != 1:
            msg = "grad_and_value only applies to real scalar-output functions."
            raise TypeError(msg)
        return vjp(vspace(val).ones()), val

    return grad_and_value_func


def vjp_and_value(func: ScalarFunction) -> VectorJacobianProductFunction:
    """Makes a function that returns vector-Jacobian-product and value of a function.

    For a vector-valued function `fun` the vector-Jacobian-product (VJP) is here
    defined as a function of a vector `v` corresponding to

        vjp(v) = v @ j

    where `j` is the Jacobian of `f = fun(x)` wrt `x` i.e. the rank-2
    tensor of first-order partial derivatives of the vector-valued function,
    such that

        j[i, k] = ∂f[i] / ∂x[k]
    """

    def vjp_and_value_func(x: ArrayLike) -> VectorJacobianProduct:
        return make_vjp(func, x)

    return vjp_and_value_func


def jacobian_and_value(func: ArrayFunction) -> JacobianFunction:
    """Makes a function that returns both the Jacobian and value of a function."""

    def jacobian_and_value_func(x: ArrayLike) -> tuple[ArrayLike, ArrayLike]:
        vjp, val = make_vjp(func, x)
        val_vspace = vspace(val)
        jacobian_shape = val_vspace.shape + vspace(x).shape
        jacobian_rows = map(vjp, val_vspace.standard_basis())
        return np.reshape(np.stack(jacobian_rows), jacobian_shape), val

    return jacobian_and_value_func


def mhp_jacobian_and_value(func: ArrayFunction) -> MatrixHessianProductFunction:
    """Makes a function that returns MHP, Jacobian and value of a function.

    For a vector-valued function `fun` the matrix-Hessian-product (MHP) is here
    defined as a function of a matrix `m` corresponding to

        mhp(m) = sum(m[:, :, None] * h[:, :, :], axis=(0, 1))

    where `h` is the vector-Hessian of `f = fun(x)` wrt `x` i.e. the rank-3
    tensor of second-order partial derivatives of the vector-valued function,
    such that

        h[i, j, k] = ∂²f[i] / (∂x[j] ∂x[k])
    """

    def mhp_jacobian_and_value_func(
        x: ArrayLike,
    ) -> tuple[MatrixHessianProduct, ArrayLike, ArrayLike]:
        mhp, (jacob, val) = make_vjp(lambda x: atuple(jacobian_and_value(func)(x)), x)
        return lambda m: mhp((m, vspace(val).zeros())), jacob, val

    return mhp_jacobian_and_value_func


def hessian_grad_and_value(func: ArrayFunction) -> HessianFunction:
    """Makes a function that returns the Hessian, gradient & value of a function."""

    def grad_func(x: ArrayLike) -> tuple[ArrayLike, ScalarLike]:
        vjp, val = make_vjp(func, x)
        return vjp(vspace(val).ones()), val

    def hessian_grad_and_value_func(
        x: ArrayLike,
    ) -> tuple[ArrayLike, ArrayLike, ScalarLike]:
        x_vspace = vspace(x)
        vjp_grad, (grad, val) = make_vjp(lambda x: atuple(grad_func(x)), x)
        hessian_shape = x_vspace.shape + x_vspace.shape
        zeros = vspace(val).zeros()
        hessian_rows = (vjp_grad((v, zeros)) for v in x_vspace.standard_basis())
        return np.reshape(np.stack(hessian_rows), hessian_shape), grad, val

    return hessian_grad_and_value_func


def mtp_hessian_grad_and_value(func: ArrayFunction) -> MatrixTressianProductFunction:
    """Makes a function that returns MTP, Jacobian and value of a function.

    For a scalar-valued function `fun` the matrix-Tressian-product (MTP) is
    here defined as a function of a matrix `m` corresponding to

        mtp(m) = sum(m[:, :] * t[:, :, :], axis=(-1, -2))

    where `t` is the 'Tressian' of `f = fun(x)` wrt `x` i.e. the 3D array of
    third-order partial derivatives of the scalar-valued function such that

        t[i, j, k] = ∂³f / (∂x[i] ∂x[j] ∂x[k])
    """

    def mtp_hessian_grad_and_value_func(
        x: ArrayLike,
    ) -> tuple[MatrixTressianProduct, ArrayLike, ArrayLike, ScalarLike]:
        mtp, (hessian, grad, val) = make_vjp(
            lambda x: atuple(hessian_grad_and_value(func)(x)),
            x,
        )
        return (
            lambda m: mtp((m, vspace(grad).zeros(), vspace(val).zeros())),
            hessian,
            grad,
            val,
        )

    return mtp_hessian_grad_and_value_func
